==> debezium-connector-spanner-2.3.6.Final-sources.jar: io/debezium/connector/spanner/kafka/internal/TaskSyncEventListener.java <==
/*
 * Copyright Debezium Authors.
 *
 * Licensed under the Apache Software License version 2.0, available at http://www.apache.org/licenses/LICENSE-2.0
 */
package io.debezium.connector.spanner.kafka.internal;

import static io.debezium.connector.spanner.task.LoggerUtils.debug;
import static org.slf4j.LoggerFactory.getLogger;

import java.time.Duration;
import java.util.ArrayList;
import java.util.List;
import java.util.function.BiConsumer;

import org.apache.kafka.clients.consumer.Consumer;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.clients.consumer.ConsumerRecords;
import org.apache.kafka.common.TopicPartition;
import org.slf4j.Logger;

import com.google.protobuf.InvalidProtocolBufferException;

import io.debezium.connector.spanner.exception.SpannerConnectorException;
import io.debezium.connector.spanner.function.BlockingBiConsumer;
import io.debezium.connector.spanner.kafka.event.proto.SyncEventProtos;
import io.debezium.connector.spanner.kafka.internal.model.SyncEventMetadata;
import io.debezium.connector.spanner.kafka.internal.model.TaskSyncEvent;
import io.debezium.connector.spanner.kafka.internal.proto.SyncEventFromProtoMapper;

/** Consumes messages from the Sync Topic */
public class TaskSyncEventListener {
    private static final Logger LOGGER = getLogger(TaskSyncEventListener.class);
    private final String consumerGroup;
    private final String topic;
    private final boolean seekBackToPreviousEpoch;
    private final Duration pollDuration;
    private final Duration commitOffsetsTimeout;
    private final long commitOffsetsInterval;
    private final SyncEventConsumerFactory<String, byte[]> consumerFactory;
    private final List<BlockingBiConsumer<TaskSyncEvent, SyncEventMetadata>> eventConsumers = new ArrayList<>();
    private final java.util.function.Consumer<RuntimeException> errorHandler;

    private volatile Thread thread;

    public TaskSyncEventListener(
                                 String consumerGroup,
                                 String topic,
                                 SyncEventConsumerFactory<String, byte[]> consumerFactory,
                                 boolean seekBackToPreviousEpoch,
                                 java.util.function.Consumer<RuntimeException> errorHandler) {

        this.consumerGroup = consumerGroup;
        this.topic = topic;
        this.seekBackToPreviousEpoch = seekBackToPreviousEpoch;
        this.pollDuration = Duration.ofMillis(consumerFactory.getConfig().syncPollDuration());
        this.commitOffsetsTimeout = Duration.ofMillis(consumerFactory.getConfig().syncCommitOffsetsTimeout());
        this.commitOffsetsInterval = consumerFactory.getConfig().syncCommitOffsetsInterval();
        this.consumerFactory = consumerFactory;
        this.errorHandler = errorHandler;
    }

    public void subscribe(BlockingBiConsumer<TaskSyncEvent, SyncEventMetadata> eventConsumer) {
        eventConsumers.add(eventConsumer);
    }

    public void unsubscribe(BiConsumer<TaskSyncEvent, SyncEventMetadata> eventConsumer) {
        eventConsumers.remove(eventConsumer);
    }

    public void start() throws InterruptedException {
        // or take all partition list and sub to them

        TopicPartition topicPartition = new TopicPartition(topic, 0);
        List<TopicPartition> assignment = List.of(topicPartition);

        Consumer<String, byte[]> consumer = consumerFactory.createConsumer(consumerGroup);
        consumer.assign(assignment);

        Long endOffset = consumer.endOffsets(assignment).get(topicPartition);
        Long beginOffset = consumer.beginningOffsets(assignment).get(topicPartition);

        long startOffset = Math.max(endOffset - 1, beginOffset);

        try {

            if (endOffset == startOffset) {
                LOGGER.info("task {}, listen: Sync topic is empty, so initial sync is finished", consumerGroup);
                for (BlockingBiConsumer<TaskSyncEvent, SyncEventMetadata> eventConsumer : eventConsumers) {
                    eventConsumer.accept(
                            null, SyncEventMetadata.builder().canInitiateRebalancing(true).build());
                }
            }
            else {
                LOGGER.info("Task {}, listen: read last message", consumerGroup);
                try {
                    consumer.seek(topicPartition, startOffset);
                    seekBackToPreviousEpoch(consumer, topicPartition, beginOffset);
                }
                catch (org.apache.kafka.common.errors.InterruptException e) {
                    throw new InterruptedException();
                }
                catch (Exception e) {
                    errorHandler.accept(
                            new SpannerConnectorException("Error during seek back the Sync Topic", e));
                    return;
                }
            }

        }
        catch (Exception ex) {
            shutdownConsumer(consumer);
            throw ex;
        }

        thread = new Thread(
                () -> {
                    try {
                        long commitOffsetStart = System.currentTimeMillis();
                        while (!Thread.currentThread().isInterrupted()) {
                            try {
                                poll(consumer, endOffset);
                                if (!consumerFactory.isAutoCommitEnabled()
                                        && commitOffsetStart + commitOffsetsInterval < System.currentTimeMillis()) {

                                    consumer.commitSync(commitOffsetsTimeout);
                                    commitOffsetStart = System.currentTimeMillis();
                                }
                            }
                            catch (org.apache.kafka.common.errors.InterruptException
                                    | InterruptedException ex) {
                                return;
                            }
                            catch (Exception e) {
                                errorHandler.accept(
                                        new SpannerConnectorException("Error during poll from the Sync Topic", e));
                                return;
                            }
                        }

                    }
                    finally {
                        shutdownConsumer(consumer);
                    }
                },
                "SpannerConnector-TaskSyncEventListener");

        thread.start();
    }

    private int poll(Consumer<String, byte[]> consumer, long endOffset)
            throws InvalidProtocolBufferException, InterruptedException {

        ConsumerRecords<String, byte[]> records = consumer.poll(pollDuration);
        LOGGER.trace("listen: poll messages count: {}", records.count());

        if (records.isEmpty()) {
            return 0;
        }

        for (ConsumerRecord<String, byte[]> record : records) {

            TaskSyncEvent taskSyncEvent = parseSyncEvent(record);
            debug(LOGGER, "Receive SyncEvent from Kafka topic: {}", taskSyncEvent);

            for (BlockingBiConsumer<TaskSyncEvent, SyncEventMetadata> eventConsumer : eventConsumers) {
                boolean canInitiateRebalancing = (record.offset() >= endOffset - 1);
                eventConsumer.accept(
                        taskSyncEvent,
                        SyncEventMetadata.builder()
                                .offset(record.offset())
                                // Once we have consumed all the messages present in the sync topic at the
                                // start of the connector, we can then connect to the rebalance topic.
                                .canInitiateRebalancing(canInitiateRebalancing)
                                .build());
            }
        }
        return records.count();
    }

    private void seekBackToPreviousEpoch(
                                         Consumer<String, byte[]> consumer, TopicPartition topicPartition, long beginOffset)
            throws InvalidProtocolBufferException {
        if (!seekBackToPreviousEpoch) {
            return;
        }
        ConsumerRecords<String, byte[]> records = consumer.poll(pollDuration);

        if (records.isEmpty()) {
            LOGGER.warn("listen: fail to poll last message");
            return;
        }

        ConsumerRecord<String, byte[]> lastRecord = records.iterator().next();
        TaskSyncEvent taskSyncEvent = parseSyncEvent(lastRecord);

        long previousEpochOffset = taskSyncEvent.getEpochOffset();
        long startOffset = Math.max(previousEpochOffset, beginOffset);

        LOGGER.info("Task {}, listen: seek back to previous epoch offset: {}", consumerGroup, startOffset);
        consumer.seek(topicPartition, startOffset);
    }

    private TaskSyncEvent parseSyncEvent(ConsumerRecord<String, byte[]> record)
            throws InvalidProtocolBufferException {
        return SyncEventFromProtoMapper.mapFromProto(
                SyncEventProtos.SyncEvent.parseFrom(record.value()));
    }

    private void shutdownConsumer(Consumer<String, byte[]> consumer) {
        try {
            consumer.unsubscribe();
            consumer.close();
        }
        catch (org.apache.kafka.common.errors.InterruptException e) {
            if (!Thread.currentThread().isInterrupted()) {
                Thread.currentThread().interrupt();
            }
        }
    }

    public void shutdown() {
        if (thread == null) {
            return;
        }
        thread.interrupt();

        while (!thread.getState().equals(Thread.State.TERMINATED)) {
        }
        thread = null;
    }
}

==> debezium-connector-spanner-2.4.2.Final-sources.jar: io/debezium/connector/spanner/kafka/internal/TaskSyncEventListener.java <==
/*
 * Copyright Debezium Authors.
 *
 * Licensed under the Apache Software License version 2.0, available at http://www.apache.org/licenses/LICENSE-2.0
 */
package io.debezium.connector.spanner.kafka.internal;

import static io.debezium.connector.spanner.task.LoggerUtils.debug;
import static org.slf4j.LoggerFactory.getLogger;

import java.time.Duration;
import java.util.ArrayList;
import java.util.List;
import java.util.function.BiConsumer;

import org.apache.kafka.clients.consumer.Consumer;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.clients.consumer.ConsumerRecords;
import org.apache.kafka.common.TopicPartition;
import org.slf4j.Logger;

import com.google.protobuf.InvalidProtocolBufferException;

import io.debezium.DebeziumException;
import io.debezium.connector.spanner.exception.SpannerConnectorException;
import io.debezium.connector.spanner.function.BlockingBiConsumer;
import io.debezium.connector.spanner.kafka.event.proto.SyncEventProtos;
import io.debezium.connector.spanner.kafka.internal.model.SyncEventMetadata;
import io.debezium.connector.spanner.kafka.internal.model.TaskSyncEvent;
import io.debezium.connector.spanner.kafka.internal.proto.SyncEventFromProtoMapper;
import io.debezium.util.Clock;
import io.debezium.util.Metronome;
import io.debezium.util.Stopwatch;

/** Consumes messages from the Sync Topic */
public class TaskSyncEventListener {
    private static final Logger LOGGER = getLogger(TaskSyncEventListener.class);
    private final String consumerGroup;
    private final String topic;
    private final boolean seekBackToPreviousEpoch;
    private final Duration pollDuration;
    private final Duration commitOffsetsTimeout;
    private final long commitOffsetsInterval;
    private final SyncEventConsumerFactory<String, byte[]> consumerFactory;
    private final List<BlockingBiConsumer<TaskSyncEvent, SyncEventMetadata>> eventConsumers = new ArrayList<>();
    private final java.util.function.Consumer<RuntimeException> errorHandler;
    private final Duration pollInterval = Duration.ofMillis(300000);
    private final Clock clock;

    private volatile Thread thread;

    public TaskSyncEventListener(
                                 String consumerGroup,
                                 String topic,
                                 SyncEventConsumerFactory<String, byte[]> consumerFactory,
                                 boolean seekBackToPreviousEpoch,
                                 java.util.function.Consumer<RuntimeException> errorHandler) {

        this.consumerGroup = consumerGroup;
        this.topic = topic;
        this.seekBackToPreviousEpoch = seekBackToPreviousEpoch;
        this.pollDuration = Duration.ofMillis(consumerFactory.getConfig().syncPollDuration());
        this.commitOffsetsTimeout = Duration.ofMillis(consumerFactory.getConfig().syncCommitOffsetsTimeout());
        this.commitOffsetsInterval = consumerFactory.getConfig().syncCommitOffsetsInterval();
        this.consumerFactory = consumerFactory;
        this.errorHandler = errorHandler;
        this.clock = Clock.system();
    }

    public void subscribe(BlockingBiConsumer<TaskSyncEvent, SyncEventMetadata> eventConsumer) {
        eventConsumers.add(eventConsumer);
    }

    public void unsubscribe(BiConsumer<TaskSyncEvent, SyncEventMetadata> eventConsumer) {
        eventConsumers.remove(eventConsumer);
    }

    public void start() throws InterruptedException {
        // or take all partition list and sub to them

        TopicPartition topicPartition = new TopicPartition(topic, 0);
        List<TopicPartition> assignment = List.of(topicPartition);

        Consumer<String, byte[]> consumer = consumerFactory.createConsumer(consumerGroup);
        consumer.assign(assignment);

        Long endOffset = consumer.endOffsets(assignment).get(topicPartition);
        Long beginOffset = consumer.beginningOffsets(assignment).get(topicPartition);

        long startOffset = Math.max(endOffset - 1, beginOffset);

        try {

            if (endOffset == startOffset) {
                LOGGER.info("task {}, listen: Sync topic is empty, so initial sync is finished", consumerGroup);
                for (BlockingBiConsumer<TaskSyncEvent, SyncEventMetadata> eventConsumer : eventConsumers) {
                    eventConsumer.accept(
                            null, SyncEventMetadata.builder().canInitiateRebalancing(true).build());
                }
            }
            else {
                LOGGER.info("Task {}, listen: read last message with start offset {} and end offset {}", consumerGroup, startOffset, endOffset);
                try {
                    LOGGER.info("Task {}, seeking back to end offset {}", consumerGroup, endOffset);
                    consumer.seek(topicPartition, startOffset);
                    LOGGER.info("Task {}, seeking back to previous epoch", consumerGroup);
                    seekBackToPreviousEpoch(consumer, topicPartition, beginOffset, startOffset);
                }
                catch (org.apache.kafka.common.errors.InterruptException e) {
                    LOGGER.info("Task {}, caught interrupt exception during reading the sync topic", consumerGroup, e);
                    throw new InterruptedException();
                }
                catch (Exception e) {
                    LOGGER.info("Task {}, Error during seek back the Sync Topic {}", consumerGroup);
                    errorHandler.accept(
                            new SpannerConnectorException("Error during seek back the Sync Topic", e));
                    return;
                }
            }

        }
        catch (Exception ex) {
            LOGGER.info("Shutdown consumer {} for ex {}", consumerGroup, ex);
            shutdownConsumer(consumer);
            errorHandler.accept(new RuntimeException(ex));
        }

        thread = new Thread(
                () -> {
                    try {
                        long commitOffsetStart = System.currentTimeMillis();
                        LOGGER.info("Task {}, beginning to poll the sync topic", consumerGroup);
                        Stopwatch sw = Stopwatch.accumulating().start();
                        while (!Thread.currentThread().isInterrupted()) {
                            try {
                                final Duration totalDuration = sw.stop().durations().statistics().getTotal();
                                if (totalDuration.toMillis() >= pollInterval.toMillis()) {
                                    // Restart the stopwatch.
                                    LOGGER.info("Task {}, still polling the sync topic", consumerGroup);
                                    sw = Stopwatch.accumulating().start();
                                }
                                else {
                                    sw.start();
                                }
                                poll(consumer, endOffset);
                                if (!consumerFactory.isAutoCommitEnabled()
                                        && commitOffsetStart + commitOffsetsInterval < System.currentTimeMillis()) {

                                    consumer.commitSync(commitOffsetsTimeout);
                                    commitOffsetStart = System.currentTimeMillis();
                                }
                            }
                            catch (org.apache.kafka.common.errors.InterruptException
                                    | InterruptedException ex) {
                                LOGGER.error("TaskSyncEventListener, caught interrupt exception {}", consumerGroup, ex);
                                Thread.currentThread().interrupt();
                                return;
                            }
                            catch (Exception e) {
                                errorHandler.accept(
                                        new SpannerConnectorException("Error during poll from the Sync Topic", e));
                                return;
                            }
                        }
                    }
                    finally {
                        shutdownConsumer(consumer);
                    }
                },
                "SpannerConnector-TaskSyncEventListener");
        thread.setUncaughtExceptionHandler((t, ex) -> {
            LOGGER.error("Error in SpannerConnector-TaskSyncEventListener, task {}, ex {}", consumerGroup, ex.getStackTrace());
            errorHandler.accept(new RuntimeException(ex));
        });

        thread.start();

    }

    private int poll(Consumer<String, byte[]> consumer, long endOffset)
            throws InvalidProtocolBufferException, InterruptedException {

        ConsumerRecords<String, byte[]> records = consumer.poll(pollDuration);
        LOGGER.trace("listen: poll messages count: {}", records.count());

        if (records.isEmpty()) {
            return 0;
        }
        for (ConsumerRecord<String, byte[]> record : records) {
            TaskSyncEvent taskSyncEvent = parseSyncEvent(record);
            debug(LOGGER, "Receive SyncEvent from Kafka topic: {}", taskSyncEvent);

            if (record.offset() == endOffset - 1) {
                LOGGER.info("Task {}, can begin to initiate rebalancing", consumerGroup);
            }
            for (BlockingBiConsumer<TaskSyncEvent, SyncEventMetadata> eventConsumer : eventConsumers) {
                boolean canInitiateRebalancing = (record.offset() >= endOffset - 1);
                eventConsumer.accept(
                        taskSyncEvent,
                        SyncEventMetadata.builder()
                                .offset(record.offset())
                                // Once we have consumed all the messages present in the sync topic at the
                                // start of the connector, we can then connect to the rebalance topic.
                                .canInitiateRebalancing(canInitiateRebalancing)
                                .build());
            }
        }
        return records.count();
    }

    private void seekBackToPreviousEpoch(
                                         Consumer<String, byte[]> consumer, TopicPartition topicPartition, long beginOffset, long currOffset)
            throws InvalidProtocolBufferException {
        if (!seekBackToPreviousEpoch) {
            LOGGER.info("Task {}, not seeking back to previous epoch");
            return;
        }
        ConsumerRecords<String, byte[]> records = consumer.poll(pollDuration);

        long currentOffset = currOffset;
        while (records.isEmpty()) {
            currentOffset = currentOffset - 1;
            if (currOffset - currentOffset >= 100) {
                throw new DebeziumException("Task " + consumerGroup + "failed to poll last message from the sync topic");
            }
            LOGGER.warn("Task {}, listen: fail to poll last message, trying again", consumerGroup);
            consumer.seek(topicPartition, currentOffset);
            records = consumer.poll(pollDuration);
        }

        ConsumerRecord<String, byte[]> lastRecord = records.iterator().next();
        TaskSyncEvent taskSyncEvent = parseSyncEvent(lastRecord);

        long previousEpochOffset = taskSyncEvent.getEpochOffset();
        long startOffset = Math.max(previousEpochOffset, beginOffset);
        LOGGER.info("Task {}, listen: found previous epoch offset {} and begin offset {}", consumerGroup, previousEpochOffset, beginOffset);

        LOGGER.info("Task {}, listen: seek back to previous epoch offset: {}", consumerGroup, startOffset);
        consumer.seek(topicPartition, startOffset);
    }

    private TaskSyncEvent parseSyncEvent(ConsumerRecord<String, byte[]> record)
            throws InvalidProtocolBufferException {
        return SyncEventFromProtoMapper.mapFromProto(
                SyncEventProtos.SyncEvent.parseFrom(record.value()));
    }

    private void shutdownConsumer(Consumer<String, byte[]> consumer) {
        try {
            LOGGER.info("TaskSyncEventListener, Lost connectivity {}", consumerGroup);
            errorHandler.accept(new SpannerConnectorException("Lost connectivity to the sync topic"));
            LOGGER.info("TaskSyncEventListener, Shutting down consumer {}", consumerGroup);
            consumer.unsubscribe();
            consumer.close();
        }
        catch (org.apache.kafka.common.errors.InterruptException e) {
            if (!Thread.currentThread().isInterrupted()) {
                Thread.currentThread().interrupt();
            }
        }
    }

    public void shutdown() {
        LOGGER.info(
                "Stopping TaskSyncEventListener for Task Uid {}",
                consumerGroup);
        if (thread == null) {
            return;
        }
        thread.interrupt();
        final Metronome metronome = Metronome.sleeper(Duration.ofMillis(100), clock);
        while (!thread.getState().equals(Thread.State.TERMINATED)) {
            try {
                // Sleep for sleepInterval.
                LOGGER.info(
                        "Still stopping TaskSyncEventListener for Task Uid {} with state {}",
                        consumerGroup, thread.getState());
                thread.interrupt();

                metronome.pause();
            }
            catch (InterruptedException e) {
                Thread.currentThread().interrupt();
            }
        }
        LOGGER.info(
                "Stopped TaskSyncEventListener for Task Uid {}",
                consumerGroup);
        thread = null;
    }
}
